
/******************************************************************************
 *
 * Copyright:
 *    (C) 2000 - 2006 Embedded Artists AB
 *
 * Description:
 *    Sample application that implements ICMP (Ping) functionality.
 *
 *****************************************************************************/

/******************************************************************************
 * Includes
 *****************************************************************************/

#include "general.h"
#include "ethapi.h"

#include <printf_P.h>
#include <string.h>
#include <lpc2xxx.h>


/******************************************************************************
 * Defines
 *****************************************************************************/
#define INBUF_LEN 250

#define ARP_REQUEST    1         /* Operation ARP request      */
#define ARP_REPLY      2         /* Operation ARP reply        */

#define CRYSTAL_FREQUENCY FOSC
#define PLL_FACTOR        PLL_MUL
#define VPBDIV_FACTOR     PBSD

#define SPI_SLAVE_CS 0x00004000  //pin P0.14
//#define ENC_RESET    0x00001000  //pin P0.12
#define FAILSAFE_VALUE 5000


extern tU16 pings;

/******************************************************************************
 * Local variables
 *****************************************************************************/

/*
 * The IP address assigned to the local host. 
 */ 
static tU8 localIP[4] = {172,17,1,230};

/*
 * The Ethernet address (MAC) assigned to the local host
 */
static tU8 mac[6] = {0x00, 0x1a, 0xf1, 0x10, 0x11, 0x12};
static tU8 inBuf[INBUF_LEN];
static tU8 bank = 0;

tU8 encReadReg(tU8 regNo);

#define  SPI_WAIT delayMs(1)


/*****************************************************************************
 *
 * Description:
 *    Delay execution by a specified number of milliseconds by using
 *    timer #1. A polled implementation.
 *
 * Params:
 *    [in] delayInMs - the number of milliseconds to delay.
 *
 ****************************************************************************/
static void
delayMs(tU16 delayInMs)
{
  /*
   * setup timer #1 for delay
   */
  TIMER1_TCR = 0x02;          //stop and reset timer
  TIMER1_PR  = 0x00;          //set prescaler to zero
  TIMER1_MR0 = delayInMs * ((CRYSTAL_FREQUENCY * PLL_FACTOR)/ (1000 * VPBDIV_FACTOR));
  TIMER1_IR  = 0xff;          //reset all interrrupt flags
  TIMER1_MCR = 0x04;          //stop timer on match
  TIMER1_TCR = 0x01;          //start timer
  
  //wait until delay time has elapsed
  while (TIMER1_TCR & 0x01)
    ;
}

/*****************************************************************************
 *
 * Description:
 *    Sends and received one byte over the SPI serial channel.
 *    A polled implementation.
 *
 * Params:
 *    [in] byte - the byte to send.
 *
 * Return:
 *    The received byte.
 *
 ****************************************************************************/
static tU8
sendSpi(tU8 byte)
{
    tU32 failsafe;
    tU8  receivedByte;
    
  SPI_SPDR = byte;
  failsafe = 0;
  while(((SPI_SPSR & 0x80) == 0) && (failsafe < FAILSAFE_VALUE))
    failsafe++;

  receivedByte = SPI_SPDR;

  //reinitialize SPI if timeout
  if (failsafe >= FAILSAFE_VALUE)
  {
    printf("\nSPI-error: SSEL signal must be high!\n");
    receivedByte = 0x00;
    SPI_SPCCR = 0x08;    
    SPI_SPCR  = 0x20;  //no irq, MSB first, master, cpol=0, cpha=0
  }
  
  return receivedByte;
}

/*****************************************************************************
 *
 * Description:
 *    Low-level routine to write a register in ENC28J60
 *
 ****************************************************************************/
void
encWriteReg(tU8 regNo, tU8 data)
{
  IOCLR0 = SPI_SLAVE_CS;  //activate SPI
  SPI_WAIT;
  sendSpi(0x40 | regNo);  //write in regNo
  sendSpi(data);
  IOSET0 = SPI_SLAVE_CS;  //deactivate SP
  SPI_WAIT;
  
  IOCLR0 = SPI_SLAVE_CS;  //activate SPI
  SPI_WAIT;
  sendSpi(0x1f);          //read reg 0x1f

  IOSET0 = SPI_SLAVE_CS;  //deactivate SP
  SPI_WAIT;
}

void
encWriteRegRetry(tU8 regNo, tU8 data, tU8 cnt)
{
  while(cnt-- > 0)
  {
    IOCLR0 = SPI_SLAVE_CS;  //activate SPI
  SPI_WAIT;
    sendSpi(0x40 | regNo);  //write in regNo
    sendSpi(data);
    IOSET0 = SPI_SLAVE_CS;  //deactivate SP
  SPI_WAIT;
  
    IOCLR0 = SPI_SLAVE_CS;  //activate SPI
  SPI_WAIT;
    sendSpi(0x1f);          //read reg 0x1f

  SPI_WAIT;
    IOSET0 = SPI_SLAVE_CS;  //deactivate SP
    
    if (data == encReadReg(regNo))
      break;
  }
  if (cnt == 0)
    printf("\nERROR too many retries...\n");
}

/*****************************************************************************
 *
 * Description:
 *    Low-level routine to read a register in ENC28J60
 *
 ****************************************************************************/
tU8
encReadReg(tU8 regNo)
{
  tU8 rxByte;

  IOCLR0 = SPI_SLAVE_CS;  //activate SPI
  SPI_WAIT;
  sendSpi(0x1f);          //read reg 0x1f

  bank = sendSpi(0) & 0x3;
  IOSET0 = SPI_SLAVE_CS;  //deactivate SP
  SPI_WAIT;


  IOCLR0 = SPI_SLAVE_CS;  //activate SPI
  SPI_WAIT;
  sendSpi(regNo);         //read reg i
  rxByte = sendSpi(0);

  //check if MAC or MII register
  if (((bank == 2) && (regNo <= 0x1a)) ||
      ((bank == 3) && (regNo <= 0x05 || regNo == 0x0a)))
    //ignore first byte and read another byte
{
    rxByte = sendSpi(0);
printf("\nRead MAC or MII reg = %x value=%x", regNo, rxByte);
}  
  IOSET0 = SPI_SLAVE_CS;  //deactivate SP
  SPI_WAIT;

  return rxByte;
}


/*****************************************************************************
 *
 * Description:
 *    Low-level routine to read a received frame in the ENC28J60
 *
 ****************************************************************************/
void
encReadBuff(tU16 length, tU8 *pBuff)
{
  IOCLR0 = SPI_SLAVE_CS;  //activate SPI
  SPI_WAIT;
  sendSpi(0x20 | 0x1a);   //read buffer memory

  while(length--)
  {
    if(pBuff != NULL)
      *pBuff++ = sendSpi(0);
    else
      sendSpi(0);
  }

  IOSET0 = SPI_SLAVE_CS;  //deactivate SP
  SPI_WAIT;
}


/*****************************************************************************
 *
 * Description:
 *    Low-level routine to write a frame to be transmitted in the ENC28J60
 *
 ****************************************************************************/
void
encWriteBuff(tU16 length, tU8 *pBuff)
{
  IOCLR0 = SPI_SLAVE_CS;  //activate SPI
  SPI_WAIT;
  sendSpi(0x60 | 0x1a);   //write buffer memory

  while(length--)
    sendSpi(*pBuff++);

  IOSET0 = SPI_SLAVE_CS;  //deactivate SP
  SPI_WAIT;
}


/*****************************************************************************
 *
 * Description:
 *    Low-level routine to set a bit in one register in the ENC28J60
 *
 ****************************************************************************/
void
encBitSet(tU8 regNo, tU8 data)
{
  IOCLR0 = SPI_SLAVE_CS;  //activate SPI
  SPI_WAIT;
  sendSpi(0x80 | regNo);  //bit field set 
  sendSpi(data);
  IOSET0 = SPI_SLAVE_CS;  //deactivate SP
  SPI_WAIT;
}


/*****************************************************************************
 *
 * Description:
 *    Low-level routine to clear a bit in one register in the ENC28J60
 *
 ****************************************************************************/
void
encBitClr(tU8 regNo, tU8 data)
{
  IOCLR0 = SPI_SLAVE_CS;  //activate SPI
  SPI_WAIT;
  sendSpi(0xA0 | regNo);  //bit field clear 
  sendSpi(data);
  IOSET0 = SPI_SLAVE_CS;  //deactivate SP
  SPI_WAIT;
}


/*****************************************************************************
 *
 * Description:
 *    Low-level routine to reset the ENC28J60
 *
 ****************************************************************************/
void
encReset(void)
{
  IOCLR0 = SPI_SLAVE_CS;  //activate SPI
  SPI_WAIT;
  sendSpi(0xff);          //soft reset
  IOSET0 = SPI_SLAVE_CS;  //deactivate SP
  SPI_WAIT;
}


/*****************************************************************************
 *
 * Description:
 *    Initialize the SPI channle #0
 *
 ****************************************************************************/
void
initSpi(void)
{
  //connect SPI bus to IO-pins
  PINSEL0 |= 0x00001500;
  
  //initialize SPI interface
  SPI_SPCCR = 0x08;    
  SPI_SPCR  = 0x20;  //no irq, MSB first, master, cpol=0, cpha=0

  //make SPI slave chip select an output and set signal high
  IODIR |= SPI_SLAVE_CS;
  IOSET  = SPI_SLAVE_CS;

#if 0
  //reset ENC28J60
  IODIR |= ENC_RESET;
  IOSET  = ENC_RESET;
  IOCLR  = ENC_RESET;
  delayMs(1);              //reset pulse width > 2 us
  IOSET  = ENC_RESET;
  delayMs(1);              //wait >300us after reset
#endif
}


/*****************************************************************************
 *
 * Description:
 *    Checksum calculation. 16-bit one's complement. 
 *
 * Params:
 *    [in] pStartAddress - pointer to beginning of data 
 *    [in] len           - length of data 
 *    [in] initialValue  - precalculated checksum (e.g. pseudo header) 
 *
 * Returns:
 *    calculated checksum 
 *
 ****************************************************************************/
static tU16
checksum(tU8* pStartAddress,
         tU16 len,
         tU16 initialValue)
{
  tU32 csum       = ~initialValue & 0x0000ffff;
  tU8 *p8bit      = NULL;
  tU16 val16      = 0;

  tU32 aWordCount = len / 2;
  p8bit           = pStartAddress;

  while (aWordCount-- > 0)
  {
    val16  = *p8bit++;
    val16 |= *p8bit << 8;
    p8bit++;
    
    csum  += val16;
  }

  if (len % 2)
  {
    /* The segment ends with odd byte, add it. */
    csum += *p8bit++;
  }

  while (csum >> 16)
  {
    csum = (csum & 0xffff) + (csum >> 16); 
    /* add in end-around carry */
  }

  return ~(tU16) csum;
}


/*****************************************************************************
 *
 * Description:
 *    Handles an incoming ICMP packet
 *
 ****************************************************************************/
static void
icmpInput(tU8* pBuf,
          tU16 len)
{

  /*
   *     +--+--+--+--+
   *  0: |ty|co|chsum|
   *     +--+--+--+--+
   * 
   * ty  = type
   * co  = code
   * 
   */

  tU16 csum = 0;

  if(len < 4)
  {
    return;
  }

  if(checksum((tU8 *)pBuf, len, 0) != 0) 
  {
    return;
  }

  if(!(pBuf[0] == 8 && pBuf[1] == 0)) 
  {
    /* Only support for Ping */
    return;
  }

pings++;

  // create reply
  pBuf[0] = 0;
  pBuf[2] = 0;
  pBuf[3] = 0;

  csum = checksum(pBuf, len, 0);

  pBuf[2] = (csum & 0xff);
  pBuf[3] = (csum >> 8);

  // create IP header

  pBuf -= 20;
  len  += 20;

  // set destination IP address (same as source in input packet)
  memcpy(&pBuf[16], &pBuf[12], 4);

  // set source IP address
  memcpy(&pBuf[12], localIP, 4);

  // checksum
  pBuf[10] = 0;
  pBuf[11] = 0;

  csum = checksum(pBuf, 20, 0);

  pBuf[10] = (csum & 0xff);
  pBuf[11] = (csum >> 8);

  // ethernet header
  pBuf -= 14;
  len  += 14;

  // destination ethernet address
  memcpy(&pBuf[0], &pBuf[6], 6);

  memcpy(&pBuf[6], mac, 6);

  ethIf_send(pBuf, len);
}


/*****************************************************************************
 *
 * Description:
 *   Handles an incoming IP packet
 *
 ****************************************************************************/
static void
ipInput(tU8* pBuf, 
        tU16 len)

{
  tU16 totLen = 0;

  /*
   *     +--+--+--+--+--+--+--+--+
   *  0: |vl|TO| len |ident|flags|
   *     +--+--+--+--+--+--+--+--+
   *  8: |TT|pr|chsum|  src IP   |
   *     +--+--+--+--+--+--+--+--+
   * 16: |  dst IP   |
   *     +--+--+--+--+
   * 
   * vl  = version and header length
   * TO  = TOS
   * len = total length
   * TT  = TTL
   * pr  = protocol
   * 
   */

  if(len < 20)
  {
    return;
  }

  // version 4, header length 20 (5 << 2)
  if(pBuf[0] != 0x45)
  {
    return;
  }

  if (checksum(pBuf, 20, 0) != 0)
  {
    return;
  }

  // only accept packets destined directly to local host
  if(memcmp(&pBuf[16], localIP, 4) != 0)
  {
    return;
  }

  totLen = ((pBuf[2] << 8) | pBuf[3]);

  // only accept ICMP
  if(pBuf[9] == 1)
    icmpInput(pBuf+20, totLen-20);
}


/*****************************************************************************
 *
 * Description:
 *    Handles an incoming ARP packet. 
 *
 ****************************************************************************/
static void
arpInput(tU8* pBuf,
         tU16 len)
{
  tBool isLocalTarg = FALSE;
  tU16  op          = 0;

  /*
   *     +--+--+--+--+--+--+--+--+
   *  0: |  ht |  pt |hs|ps| op  |
   *     +--+--+--+--+--+--+--+--+
   *  8: |  sender eth     | send
   *     +--+--+--+--+--+--+--+--+
   * 16: | IP  |  target eth     |
   *     +--+--+--+--+--+--+--+--+
   * 24: | target IP |
   *     +--+--+--+--+
   *
   *
   * ht = hardware address type
   * pt = protocol address type
   * hs = hardware address size
   * ps = protocol address size
   * op = operation
   * 
   */
printf("\narpInput");

  if(len < 28)
    return;

  op = (pBuf[6] << 8 | pBuf[7]);
  isLocalTarg = memcmp(&pBuf[24], localIP, 4) == 0;
  
  if(op == ARP_REQUEST && isLocalTarg == TRUE)
  {
    // create a reply

    // set operation to ARP reply
    pBuf[6] = ((ARP_REPLY >> 8) & 0xff);
    pBuf[7] = (ARP_REPLY & 0xff);

    // copy sender ethernet address to target ethernet address position
    memcpy(&pBuf[18], &pBuf[8], 6);

    // copy sender IP address to target IP address position
    memcpy(&pBuf[24], &pBuf[14], 4);

    // copy local mac address to sender ethernet address position
    memcpy(&pBuf[8], mac, 6);

    // copy local IP address to sender IP address position
    memcpy(&pBuf[14], localIP, 4);

    // create ethernet header
    pBuf -= 14;

    // destination ethernet address
    memcpy(&pBuf[0], &pBuf[14+18], 6);

    // local ethernet address
    memcpy(&pBuf[6], mac, 6);

    // ethernet type (= ARP, 0x0806)
    pBuf[12] = (0x0806 >> 8);
    pBuf[13] = (0x0806 & 0xff);

    ethIf_send(pBuf, 42);
  }
}


/*****************************************************************************
 *
 * Description:
 *    Handles an incoming Ethernet packet. 
 *
 ****************************************************************************/
static void
ethInput(tU8* pBuf, 
         tU16 len)
{
  /*
   * Ethernet header:
   * 
   * | dest addr | src addr | type |
   *       6          6         2      
   */

  tU16 type = 0;
printf("\nethInput");
  if(len < 14)
    return;

  type = (pBuf[12] << 8 | pBuf[13]);

  /* move data pointer past ethernet header */
  pBuf += 14;
  len  -= 14;

  if(type < 0x0800)
  {
    /* 802.2/802.3 not supported */
    return;
  }

  /* send the packet to the destination protocol */

  switch(type)
  {
  case 0x0800: /* IP packet */
    ipInput(pBuf, len);
    break;
  case 0x0806: /* ARP packet */
    arpInput(pBuf, len);
    break;
  default:
    break;
  }
}


/*****************************************************************************
 *
 * Description:
 *    
 *
 ****************************************************************************/
tU8 ethInit(void)
{
  //initialize spi and (hard) reset ENC28J60
  initSpi();

  printf("\nThe MAC address is: %x:%x:%x:%x:%x:%x", mac[0], mac[1], mac[2], mac[3], mac[4], mac[5]);
  printf("\nand the IP-address is: %d.%d.%d.%d", localIP[0], localIP[1], localIP[2], localIP[3]);
  printf("\nTry pinging the board...\n");

  // initialize driver
  return ethIf_init(mac);
}

/*****************************************************************************
 *
 * Description:
 *    
 *
 ****************************************************************************/
void ethPoll(void)
{
  tU16 len = 0;

  {
    len = ethIf_poll(inBuf, INBUF_LEN);

    if(len > 0)
    {
      ethInput(inBuf, len);
    }
  }
}
