/******************************************************************************
 *
 * Copyright:
 *    (C) 2000 - 2005 Embedded Artists AB
 *
 * Description:
 *    Ethernet driver for Davicom DM9000E
 *
 *****************************************************************************/

/******************************************************************************
 * Includes 
 *****************************************************************************/

#include "general.h"
#include "ethdrv_enc28j60.h"
#include <lpc2xxx.h>
#include <string.h>

#include <printf_P.h>

/******************************************************************************
 * Local variables
 *****************************************************************************/

/* Ethernet address (MAC) */
static tU8 macAddr[6];

static tU8 packetheader[6];
#define nextpacket_low   0
#define nextpacket_high  1
#define rec_bytecnt_low  2
#define rec_bytecnt_high 3
#define rec_status_low   4
#define rec_status_high  5

/******************************************************************************
 * Defines and typedefs
 *****************************************************************************/


#define m_eth_dis_int
#define m_eth_ena_int

#define m_nic_read(reg) encReadReg(reg)
#define m_nic_write(reg, data) encWriteReg(reg, data)
#define m_nic_write_retry(reg, data, cnt) encWriteRegRetry(reg, data, cnt)
#define m_nic_read_data(len, buf) encReadBuff((len), (buf))
#define m_nic_write_data(len, buf) encWriteBuff((len), (buf))


// bit field set
#define m_nic_bfs(reg, data) encBitSet(reg, data)

// bit field clear
#define m_nic_bfc(reg, data) encBitClr(reg, data)

/* maximum frame length */
#define ENC_MAX_FRM_LEN 1518

/*
 * Controller memory layout:
 *
 * 0x0000 - 0x0bff  3k bytes transmit buffer   //no longer true
 * 0x0c00 - 0x1fff  5k bytes receive buffer     //no longer true
 */
#define ENC_TX_BUF_LEN   ENC_MAX_FRM_LEN
#define ENC_TX_BUF_START (8192 - (ENC_TX_BUF_LEN + 8))
#define ENC_TX_BUF_STOP  (ENC_TX_BUF_START + (ENC_TX_BUF_LEN + 8))

#define ENC_RX_BUF_START 0x0000
#define ENC_RX_BUF_STOP  ((ENC_TX_BUF_START - 2) | 0x0001)
#define ENC_RX_BUF_LEN   (ENC_RX_BUF_STOP - ENC_RX_BUF_START + 1)



/******************************************************************************
 * External Function Prototypes
 *****************************************************************************/

void encWriteReg(tU8 regNo, tU8 data);
tU8  encReadReg(tU8 regNo);
void encReadBuff(tU16 length, tU8 *pBuff);
void encWriteBuff(tU16 length, tU8 *pBuff);
void encBitSet(tU8 regNo, tU8 data);
void encBitClr(tU8 regNo, tU8 data);

/******************************************************************************
 * Implementation of local functions
 *****************************************************************************/
tU8 ethIf_init(tU8* pEthAddr);


void banksel(tU16 reg)
{
  m_nic_bfc(CTL_REG_ECON1, ENC_ECON1_BSEL1 | ENC_ECON1_BSEL0);
  m_nic_bfs(CTL_REG_ECON1, (reg >> 8) & 0x03);
}

/*****************************************************************************
 *
 * Description:
 *    Read PHY registers.
 *
 *    NOTE! This function will change to Bank 2.
 *
 * Params:
 *    [in] addr address of the register to read
 *
 * Returns:
 *    The value in the register
 *
 ****************************************************************************/
static tU16 
phyRead(tU8 addr)
{
  tU16 ret = 0;

  /* write address to MIREGADR */
  banksel(CTL_REG_MIREGADR);
  m_nic_write(CTL_REG_MIREGADR, addr);

  /* set MICMD.MIIRD */
  m_nic_write(CTL_REG_MICMD, ENC_MICMD_MIIRD);

  /* poll MISTAT.BUSY bit until operation is complete */
  banksel(CTL_REG_MISTAT);
  while((m_nic_read(CTL_REG_MISTAT) & ENC_MISTAT_BUSY) != 0)
  {
    static int cnt = 0;

    if(cnt++ >= 10)
    {
      printf("#");
      cnt = 0;
      break;
    }
  }

  /* clear MICMD.MIIRD */
  banksel(CTL_REG_MICMD);
  m_nic_write(CTL_REG_MICMD, 0);

  ret  = (m_nic_read(CTL_REG_MIRDH) << 8);
  ret |= (m_nic_read(CTL_REG_MIRDL) & 0xFF);

  return ret;
}


static void 
phyWrite(tU8 addr, tU16 data)
{
  tU16 ret = 0;

  /* write address to MIREGADR */
  banksel(CTL_REG_MIREGADR);
  m_nic_write(CTL_REG_MIREGADR, addr);
  m_nic_write(CTL_REG_MIWRL, data & 0xff);
  m_nic_write(CTL_REG_MIWRH, data >> 8);

  /* poll MISTAT.BUSY bit until operation is complete */
  banksel(CTL_REG_MISTAT);
  while((m_nic_read(CTL_REG_MISTAT) & ENC_MISTAT_BUSY) != 0)
  {
    static int cnt = 0;

    if(cnt++ >= 10)
    {
      printf("#");
      cnt = 0;
      break;
    }
  }
}

/*****************************************************************************
 *
 * Description:
 *    Get the next packet from the receive ring buffer. 
 *
 * Params:
 *    [in/out] pBuf - data is copied to this buffer
 *    [in]     len  - length of buffer
 *
 * Returns:
 *    number of copied bytes to the supplied buffer
 *
 ****************************************************************************/
static tU16
getPacket(tU8* pBuf, 
          tU16 len)
{
  tU16 nextPointer;
  tU16 count = 0;
  tU16 cpyLen = 0;
//  tU16 status = 0;

  //Get Ethernet Frame from Buffer
  banksel(CTL_REG_ERDPTL);
  m_nic_write(CTL_REG_ERDPTL, packetheader[nextpacket_low]);
  m_nic_write(CTL_REG_ERDPTH, packetheader[nextpacket_high]);

  m_nic_read_data(6, packetheader);

  count = (packetheader[rec_bytecnt_high] << 8) + packetheader[rec_bytecnt_low];

  //protect from oversized packet
  if (ENC_MAX_FRM_LEN >= count)
  {
    if(pBuf != NULL)
    {
      if(count <= len)
        cpyLen = count;
      else
        cpyLen = len;
    }

    /* read data into buffer */
    m_nic_read_data(cpyLen, pBuf);

    /* skip any remaining data */
    m_nic_read_data((count-cpyLen), NULL);
  }

  nextPointer = (packetheader[nextpacket_high] << 8) + packetheader[nextpacket_low];
     
  //handle errata read pointer start (must be odd)
  nextPointer--;
       
  if ((nextPointer < ENC_RX_BUF_START) || (nextPointer > ENC_RX_BUF_STOP))
    nextPointer = ENC_RX_BUF_STOP;

  /* decrease packet counter */
  m_nic_bfs(CTL_REG_ECON2, ENC_ECON2_PKTDEC);
  
  banksel(CTL_REG_ERXRDPTL);
  m_nic_write(CTL_REG_ERXRDPTL, nextPointer & 0xff);
  m_nic_write(CTL_REG_ERXRDPTH, nextPointer >> 8);
  
  return cpyLen;
}

/******************************************************************************
 * Implementation of public functions
 *****************************************************************************/

/*****************************************************************************
 *
 * Description:
 *    Resets the NIC and initializes all required hardware registers. 
 *
 * Params:
 *    [in] pEthAddr - the ethernet address (MAC) that should be assigned to 
 *                    the driver. 
 *
 ****************************************************************************/
tU8
ethIf_init(tU8* pEthAddr)
{
  tU16 phid1 = 0;
  tU16 phid2 = 0;

  memcpy(macAddr, pEthAddr, 6);

  /* Wait for Oscillator Start-up Timer (OST). */
  while((m_nic_read(CTL_REG_ESTAT) & ENC_ESTAT_CLKRDY) == 0)
  {
    static int cnt = 0;

    if(cnt++ >= 1000)
    {
      printf(".");
      cnt = 0;
    }
  }

  /* soft reset */
  encReset();
  delayMs(5);

  /* stop send & receive */
  m_nic_bfc(CTL_REG_ECON1, ENC_ECON1_RXEN);
  m_nic_bfc(CTL_REG_ECON1, ENC_ECON1_TXRTS);

  /* auto-increment RX-pointer when reading a received packet */
  m_nic_bfs(CTL_REG_ECON2, ENC_ECON2_AUTOINC);

  /* 
   * Setup the buffer space. Only start pointer is needed since the default
   * value of the end pointer (ERXND) is 0x1fff.
   */  
  packetheader[nextpacket_low]  = (ENC_RX_BUF_START & 0xff);
  packetheader[nextpacket_high] = (ENC_RX_BUF_START >> 8);

  banksel(CTL_REG_ERDPTL);
  m_nic_write(CTL_REG_ERDPTL, (ENC_RX_BUF_START & 0xFF));
  m_nic_write(CTL_REG_ERDPTH, (ENC_RX_BUF_START >> 8));

  banksel(CTL_REG_ERXSTL);
  m_nic_write(CTL_REG_ERXSTL,   (ENC_RX_BUF_START & 0xFF));
  m_nic_write(CTL_REG_ERXSTH,   (ENC_RX_BUF_START >> 8));
  m_nic_write(CTL_REG_ERXRDPTL, (ENC_RX_BUF_STOP & 0xFF));
  m_nic_write(CTL_REG_ERXRDPTH, (ENC_RX_BUF_STOP >> 8));
  m_nic_write(CTL_REG_ERXNDL,   (ENC_RX_BUF_STOP & 0xFF));
  m_nic_write(CTL_REG_ERXNDH,   (ENC_RX_BUF_STOP >> 8));
  m_nic_write(CTL_REG_ETXSTL,   (ENC_TX_BUF_START & 0xFF));
  m_nic_write(CTL_REG_ETXSTH,   (ENC_TX_BUF_START >> 8));

  /* enable MAC to receive frames */
  banksel(CTL_REG_MACON1);
  m_nic_write_retry(CTL_REG_MACON1, ENC_MACON1_TXPAUS | ENC_MACON1_RXPAUS | ENC_MACON1_MARXEN, 10);
  m_nic_write_retry(CTL_REG_MACON3, ENC_MACON3_PADCFG0 | ENC_MACON3_TXCRCEN | ENC_MACON3_FRMLNEN, 10);

  /* don't timeout transmissions on saturated media */
  m_nic_write(CTL_REG_MACON4, (1<<6) /*ENC_MACON4_DEFER*/);

  /* collisions occur at 63rd byte */
  m_nic_write(CTL_REG_MACLCON2, 63);

  // clear MARTST
  m_nic_write_retry(CTL_REG_MACON2, 0, 10);

  m_nic_write_retry(CTL_REG_MAIPGL, 0x12, 10);
  m_nic_write_retry(CTL_REG_MAIPGH, 0x0C, 10);

  /* set maximum frame length */
  m_nic_write_retry(CTL_REG_MAMXFLL, (ENC_MAX_FRM_LEN & 0xff), 10);
  m_nic_write_retry(CTL_REG_MAMXFLH, (ENC_MAX_FRM_LEN >> 8), 10);

  /* 
   * Set MAC back-to-back inter-packet gap. Recommended 0x12 for half duplex
   * and 0x15 for full duplex.
   */
  m_nic_write_retry(CTL_REG_MABBIPG, 0x12, 10);

  /* Set (low byte) Non-Back-to_Back Inter-Packet Gap. Recommended 0x12 */
  m_nic_write_retry(CTL_REG_MAIPGL, 0x12, 10);
  
  /* 
   * Set (high byte) Non-Back-to_Back Inter-Packet Gap. Recommended 
   * 0x0c for half-duplex. Nothing for full-duplex
   */
  m_nic_write_retry(CTL_REG_MAIPGH, 0x0C, 10);


  /* set MAC address */
  banksel(CTL_REG_MAADR0);
  m_nic_write_retry(CTL_REG_MAADR0, macAddr[5], 10);
  m_nic_write_retry(CTL_REG_MAADR1, macAddr[4], 10);
  m_nic_write_retry(CTL_REG_MAADR2, macAddr[3], 10);
  m_nic_write_retry(CTL_REG_MAADR3, macAddr[2], 10);
  m_nic_write_retry(CTL_REG_MAADR4, macAddr[1], 10);
  m_nic_write_retry(CTL_REG_MAADR5, macAddr[0], 10);

  //half duplex 
  phyWrite(PHY_REG_PHCON2, (1<<8) /*PHCON2_HDLDIS*/);
  phyWrite(PHY_REG_PHCON1, 0x0000);

  //set LED options (led A = link, led B = tx/rx)
  phyWrite(PHY_REG_PHLCON, 0x0472);
   
  // OR-filtering, Unicast, CRC-check and broadcast
//  m_nic_write(CTL_REG_ERXFCON, (ENC_RFR_UCEN|ENC_RFR_CRCEN|ENC_RFR_BCEN));

  /* enable reception */
  m_nic_bfs(CTL_REG_ECON1, ENC_ECON1_RXEN);

  /* verify identification */
  phid1 = phyRead(PHY_REG_PHID1);
  phid2 = phyRead(PHY_REG_PHID2);

  if(phid1 != ENC_PHID1_VALUE 
     || (phid2 & ENC_PHID2_MASK) != ENC_PHID2_VALUE)
  {
    printf("ERROR: failed to identify controller\n");
    printf("phid1 = %x, phid2 = %x\n", phid1, (phid2&ENC_PHID2_MASK));
    printf("should be phid1 = %x, phid2 = %x\n", ENC_PHID1_VALUE, ENC_PHID2_VALUE);
    return FALSE;
  }
  else
    return TRUE;

}

/*****************************************************************************
 *
 * Description:
 *    Send an ethernet packet. 
 *
 * Params:
 *    [in] pData - the data to send
 *    [in] len   - length of the data to send
 *
 ****************************************************************************/
void
ethIf_send(tU8* pData, 
           tU16 len)
{
  tU8 byteVal = 0;
  tU8 i,j, eirval;
  
  /* set start of transmit buffer */
  banksel(CTL_REG_EWRPTL);
  m_nic_write(CTL_REG_EWRPTL, (ENC_TX_BUF_START & 0xff));
  m_nic_write(CTL_REG_EWRPTH, (ENC_TX_BUF_START >> 8));

  /* write control byte */
  byteVal = 0x0E;
  m_nic_write_data(1, &byteVal);

  /* write data */
  m_nic_write_data(len, pData);

  i=0;
  while(i<15)
  {
    banksel(CTL_REG_ETXSTL);
    m_nic_write(CTL_REG_ETXSTL, ENC_TX_BUF_START);
    m_nic_write(CTL_REG_ETXSTH, ENC_TX_BUF_START >> 8);

    banksel(CTL_REG_ETXNDL);
    m_nic_write(CTL_REG_ETXNDL, (ENC_TX_BUF_START - 1 + len));
    m_nic_write(CTL_REG_ETXNDH, (ENC_TX_BUF_START - 1 + len) >> 8);

    //Errata #10 - Reset transmit logic before send
    m_nic_bfs(CTL_REG_ECON1, ENC_ECON1_TXRST);
    m_nic_bfc(CTL_REG_ECON1, ENC_ECON1_TXRST);
  
    //B5 Errata #10 & #13: Reset interrupt error flags
    m_nic_bfc(CTL_REG_EIR, ENC_EIR_TXERIF | ENC_EIR_TXIF);
    
    //trigger send
    m_nic_bfs(CTL_REG_ECON1, ENC_ECON1_TXRTS);
  
    //fix for transmit stalls (derived from errata B5 #13), watches TXIF and TXERIF bits
    //also implements a ~3.75ms (15 * 250us) timeout if send fails (occurs on random packet collisions)
    //btw: this took over 10 hours to fix due to the elusive undocumented bug
    j = 0;
    while(j<=15)
    {
      eirval = m_nic_read(CTL_REG_EIR);
      if ((eirval & (ENC_EIR_TXERIF | ENC_EIR_TXIF)) > 0)
        break;
      if (++j >= 15)
      {
        eirval = ENC_EIR_TXERIF;
        break;
      }
      delayMs(1);  //250us
    }

    //B5 Errata #13 - Reset TXRTS if failed send then reset logic
    m_nic_bfc(CTL_REG_ECON1, ENC_ECON1_TXRTS);
  
    if ((eirval & ENC_EIR_TXERIF) == 0)
      break;  //successful send (no error interrupt)

    i++;
    delayMs(1);  //1ms
  }
  
  
  // TODO wait for ECON1.TXRTS to be cleared ????
//while ((m_nic_read(CTL_REG_ECON1) & 0x08) == 0x08)
//  printf("?");

  /* clear transmit interrupt flag */
//  m_nic_bfc(CTL_REG_EIR, ENC_EIR_TXIF);
}


/*****************************************************************************
 *
 * Description:
 *    Poll the driver
 *
 * Params:
 *    [in] pBuf - allocated buffer to which data will be copied
 *    [in] len  - length of buffer
 *
 * Returns:
 *    Number of copied bytes to the supplied buffer
 *
 ****************************************************************************/
tU16
ethIf_poll(tU8* pBuf, 
           tU16 len)
{
  tU16 recvLen = 0;
  tU8 eir = 0;

  eir = m_nic_read(CTL_REG_EIR);
//printf("\nethIf_poll: 0x%x", eir);

  /* receive error */
  if(eir & ENC_EIR_RXERIF)
  {
    printf("ethIf_poll: Receive Error\n");
    m_nic_bfc(CTL_REG_EIR, ENC_EIR_RXERIF);
  }

  /* transmit error */
  if(eir & ENC_EIR_TXERIF)
  {
    printf("ethIf_poll: Transmit Error\n");
    m_nic_bfc(CTL_REG_EIR, ENC_EIR_TXERIF);
  }

  /* wake-up on LAN */
  if(eir & ENC_EIR_WOLIF)
  {
    printf("ethIf_poll: Wake-up on LAN\n");
  }

  /* Transmit interrupt -> a packet has been sent. */
  if((eir & ENC_EIR_TXIF) != 0)
  {
    printf("ethIf_poll: Transmit done\n");
    m_nic_bfc(CTL_REG_EIR, ENC_EIR_TXIF);
  }

  /* Link change */
  if((eir & ENC_EIR_LINKIF) != 0)
  {
    printf("\nethIf_poll: Link Change\n");
    m_nic_bfc(CTL_REG_EIR, ENC_EIR_LINKIF);
  }

  /* DMA completed */
  if((eir & ENC_EIR_DMAIF) != 0)
  {
    printf("\nethIf_poll: DMA Completed\n");
    m_nic_bfc(CTL_REG_EIR, ENC_EIR_DMAIF);
  }

  /* a packet has been received */
  if((eir & ENC_EIR_PKTIF) != 0)
  {
    recvLen = getPacket(pBuf, len);
    printf("ethIf_poll: Packet received\n");
  }

  return recvLen;
}
